import { OdataMetadata, OdataEntityType, OdataProperty, OdataEntitySet, OdataNavigationProperty, OdataHelpers, OdataAction } from 'odata-context';
import * as fs from 'fs';
import * as path from 'path';
import * as jquery from 'jquery';
import * as http from 'http';
import * as jsdom from 'jsdom';
import * as co from 'co';
import * as prompt from 'co-prompt';

import { Helpers } from './templates/Helpers';
import { StaticClasses } from './templates/static-classes';
import { Service } from './templates/service';
import { OdataComplexType } from "odata-context/odata/complextype";

export interface IConfig {
    url: string;
    dir?: string;
    skipContext?: boolean;
    skipService?: boolean;
    serviceDir?: string;
    serviceName?: string;
    import?: string;
}

export class OdataParser {
    public metadata: OdataMetadata<any>;
    private _initialized: () => void;
    private _failed: (reason: any) => void;
    public onInitialize: Promise<void>;
    private url: string;
    private dir: string;
    private skipService: boolean;
    private skipContext: boolean;
    private serviceDir: string;
    private serviceName: string;
    private import: string;
    private staticClasses: StaticClasses;
    private serviceGenerator: Service;
    private imports: { name: string, namespace: string }[] = [];

    constructor(private config: IConfig) {
        this.url = config.url;
        this.dir = config.dir || './models';
        this.skipService = config.skipService || false;
        this.skipContext = config.skipContext || false;
        this.serviceDir = config.serviceDir || this.dir;
        this.serviceName = config.serviceName || 'OdataService';
        this.import = config.import || 'odata-context';

        this.staticClasses = new StaticClasses(this.import);
        this.serviceGenerator = new Service(this.serviceDir, this.dir, this.import, this.serviceName, this.url);

        this.onInitialize = new Promise<void>((resolve, reject) => {
            this._initialized = resolve;
            this._failed = reject;
        });
        this.init();
    }

    private async init() {
        let $: JQueryStatic = null;
        const promise = new Promise<JQuery>((resolve, reject) => {
            jsdom.env("", (err, window) => {
                if (err) {
                    reject(err);
                }
                $ = require("jquery")(window) as JQueryStatic;
                $.get(this.url + '/$metadata', (data) => {
                    resolve($(data));
                }).fail(reject);
            });
        });
        try {
            const metadata = await promise;
            this.metadata = new OdataMetadata(this.url, null, $, await promise);
            this._initialized();
        } catch (ex) {
            this._failed(ex);
        }
    }

    public CreateEntities() {
        if (!fs.existsSync(this.dir)) {
            fs.mkdirSync(this.dir);
            console.log('Created dir', this.dir);
        }

        fs.writeFileSync(`${this.dir}/IFilterable.ts`, this.staticClasses.IFilterable());
        fs.writeFileSync(`${this.dir}/ISelectable.ts`, this.staticClasses.ISelectable());
        fs.writeFileSync(`${this.dir}/IExpandable.ts`, this.staticClasses.IExpandable());

        let typesContent = '';
        this.metadata.entityTypes.forEach(async entityType => {
            const file = `${this.dir}/${entityType.namespace}.${entityType.name}.ts`;
            fs.writeFile(file, this.CreateEntity(entityType), err => err && console.error(err));
            typesContent += this.CreateImport(entityType.namespace, entityType.name);
            console.log('Created entity', file);
        });

        this.metadata.complexTypes.forEach(async complexType => {
            const file = `${this.dir}/${complexType.namespace}.${complexType.name}.ts`;
            fs.writeFile(file, this.CreateEntity(complexType), err => err && console.error(err));
            typesContent += this.CreateImport(complexType.namespace, complexType.name);
            console.log('Created entity', file);
        });

        this.metadata.entityTypes.forEach(async entityType => {
            typesContent += ` export type I${entityType.name} = I${entityType.name}Model & Partial<I${entityType.name}ModelNavigationProperties> & Partial<I${entityType.name}ModelActions>;\n\n`
        });

        if (!this.skipContext) {
            const file = `${this.dir}/core.ts`;
            fs.writeFile(file, this.CreateEntitySets(this.metadata.entitySets), err => err && console.error(err));
            console.log('Created context', file);
        }
        if (!this.skipService) {
            if (!fs.existsSync(this.serviceDir)) {
                fs.mkdirSync(this.serviceDir);
                console.log('Created folder', this.serviceDir);
            }
            const file = `${this.serviceDir}/${this.serviceName}.ts`;
            fs.writeFile(file, this.serviceGenerator.Generate(), err => err && console.error(err));
            console.log('Created service', file);

        }

        fs.writeFileSync(`${this.dir}/types.ts`, typesContent);
        console.log(`TypesCreated ${this.dir}/types.ts`);
    }

    private CreateSet(set: OdataEntitySet) {
        const types = Helpers.GetTypes(set.entityType.name);
        return `    ${set.name}: OdataDbSet<${types.join(', ')}>;`
    }

    public CreateEntitySets(entitySets: OdataEntitySet[]) {
        var content = `import { OdataDbSet } from '${this.import}';\n\n`;
        entitySets.forEach(entitySet => {
            const entityType = entitySet.entityType;
            content += this.CreateImport(entityType.namespace, entityType.name);
        });
        content += `export interface IOdataSets {\n${entitySets.map(set => this.CreateSet(set)).join('\n')}\n}\n`;
        return content;
    }

    private CreateProperty(entityType: OdataEntityType | OdataComplexType, property: OdataProperty, any = false, suffix = 'Model') {
        if (property.parsedType.isComplex) {
            console.log('here', property.name);
            const optional = property.isNullable || entityType.keys.includes(property.name);
            const content = `    ${property.name}${(optional || any) ? '?' : ''}: ${any ? 'any' : 'I' + property.parsedType.name + suffix};`
            return content;
        }
        const type = OdataHelpers.ParseType(property.type);
        const optional = property.isNullable || entityType.keys.includes(property.name);
        const content = `    ${property.name}${(optional || any) ? '?' : ''}: ${any ? 'any' : Helpers.GetType(type.name, type.isCollection)};`
        return content;
    }

    private CreateNavigationProperty(entityType: OdataEntityType | OdataComplexType, property: OdataNavigationProperty) {
        const type = property.parsedType;
        let propType = `I${type.name}Model & I${type.name}ModelNavigationProperties`;
        if (type.isCollection) {
            propType = `Array<${propType}>`;
        }
        const optional = true;
        const content = `   ${property.name}${optional ? '?' : ''}: ${propType};`
        return content;
    }

    private CreateOdataNavigationProperty(entityType: OdataEntityType | OdataComplexType, property: OdataNavigationProperty) {
        const type = property.parsedType;
        const optional = true;
        const content = `   ${property.name}${optional ? '?' : ''}: I${type.name}ModelProperties & I${type.name}ModelOdataNavigationProperties;`
        return content;
    }

    private CreateExpandables(entityType: OdataEntityType | OdataComplexType, property: OdataNavigationProperty) {
        const type = property.parsedType;
        const optional = true;
        const content = `   ${property.name}${optional ? '?' : ''}: I${type.name}ModelExpandables;`
        return content;
    }

    private CreateFilterables(entityType: OdataEntityType | OdataComplexType, property: OdataNavigationProperty) {
        const type = property.parsedType;
        const optional = true;
        const content = `    ${property.name}${optional ? '?' : ''}: I${type.name}ModelFilterables;`
        return content;
    }

    private CreateOrderBy(entityType: OdataEntityType | OdataComplexType, property: OdataNavigationProperty) {
        const type = property.parsedType;
        const content = `    ${property.name}?: I${type.name}ModelOrderBy;`
        return content;
    }

    public CreateKeys(keys: string[], entityType: OdataEntityType | OdataComplexType) {
        if (keys.length === 1) {
            const property = entityType.properties.find(prop => prop.name == keys[0]);
            if (property) {
                const type = OdataHelpers.ParseType(property.type);
                return `export type I${entityType.name}ModelKeys = ${Helpers.GetType(type.name, type.isCollection)};`
            } else {
                return '';
            }

        } else {
            const content = keys.map(key => {
                const property = entityType.properties.find(p => p.name === key);
                if (property) {
                    const type = OdataHelpers.ParseType(property.type);
                    return `    ${property.name}: ${Helpers.GetType(type.name, type.isCollection)};`
                } else {
                    return undefined;
                }
            });
            return `export interface I${entityType.name}ModelKeys {
                ${content}
            }`;
        }
    }

    private CreateSelectables(entityType: OdataEntityType | OdataComplexType, property: OdataNavigationProperty | OdataProperty) {
        if (property instanceof OdataNavigationProperty) {
            var type = property.parsedType;
            return `    ${property.name}?: I${type.name}ModelSelectables;`;
        } else {
            const type = OdataHelpers.ParseType(property.type);
            return `    ${property.name}?: any;`;
        }
    }

    private GetCountableProperties(entityType: OdataEntityType | OdataComplexType, property: OdataNavigationProperty) {
        if (property.parsedType.isCollection) {
            return `    '${property.name}@odata.count'?: number;`;
        } else {
            return undefined;
        }
    }

    public CreateModel(entityType: OdataEntityType | OdataComplexType) {
        let fileContent = '';
        fileContent +=
            `import { IFilterable } from './IFilterable';
import { IExpandable } from './IExpandable';
import { ISelectable } from './ISelectable';
import { IOrderByCommand } from '${this.import}';
import { Observable } from 'rxjs/Observable';

${this.CreateKeys(entityType.keys, entityType)}

export interface I${entityType.name}Model {
${entityType.properties.map(prop => this.CreateProperty(entityType, prop)).join('\n')}
${entityType.navigationproperties.map(prop => this.GetCountableProperties(entityType, prop)).join('\n')}
}

export interface I${entityType.name}ModelProperties {
${entityType.properties.map(prop => this.CreateProperty(entityType, prop, !prop.parsedType.isComplex, 'ModelProperties')).join('\n')}
}

export interface I${entityType.name}ModelNavigationProperties {
${entityType.navigationproperties.map(prop => this.CreateNavigationProperty(entityType, prop)).join('\n')}
}

export interface I${entityType.name}ModelOdataNavigationProperties {
${entityType.navigationproperties.map(prop => this.CreateOdataNavigationProperty(entityType, prop)).join('\n')}
}

export interface I${entityType.name}ModelExpandables extends IExpandable {
${entityType.navigationproperties.map(prop => this.CreateExpandables(entityType, prop)).join('\n')}
}

export interface I${entityType.name}ModelSelectables extends ISelectable {
${entityType.properties.map(prop => this.CreateSelectables(entityType, prop)).join('\n')}
${entityType.navigationproperties.map(prop => this.CreateSelectables(entityType, prop)).join('\n')}
}

export interface I${entityType.name}ModelFilterables extends IFilterable<
    I${entityType.name}ModelProperties & I${entityType.name}ModelOdataNavigationProperties,
    I${entityType.name}ModelOdataNavigationProperties,
    I${entityType.name}ModelFilterables
> {
${entityType.navigationproperties.map(prop => this.CreateFilterables(entityType, prop)).join('\n')}
}

export interface I${entityType.name}ModelOrderBy extends IOrderByCommand<
    I${entityType.name}ModelProperties & I${entityType.name}ModelOdataNavigationProperties,
    I${entityType.name}ModelOrderBy
> {
${entityType.navigationproperties.map(prop => this.CreateOrderBy(entityType, prop)).join('\n')}
}

export interface I${entityType.name}ModelActions {
    ${entityType.actions.map(action => this.GetAction(entityType, action)).join('\n    ')}
}

`;
        return fileContent;
    }

    private GetAction(entityType: OdataEntityType | OdataComplexType, action: OdataAction) {
        const params = [];
        const parameters = action.parameters.filter(param => param.name != 'bindingParameter');
        parameters.forEach(param => {
            params.push(`${param.name}${param.isNullable ? '?' : ''}: ${Helpers.GetType(param.parsedType.name, param.parsedType.isCollection)}`);
        });

        let returnType = `Observable<void>`;
        if (action.returns) {
            let returns = 'any';
            if (action.returnType instanceof OdataEntityType) {
                returns = `I${action.returnType.name}Model & Partial<I${action.returnType.name}ModelNavigationProperties> & Partial<I${action.returnType.name}ModelActions>`
            } else {
                const type = OdataHelpers.ParseType(action.returnType);
                returns = Helpers.GetType(type.name, type.isCollection);
            }
            returnType = `Observable<${returns}>`;
        }
        const content = `${action.name}(params: ${params.length ? '{' : 'any'}${params.join(', ')}${params.length ? '}' : ''}): ${returnType};`;
        return content;
    }

    private CreateImport(namespace: string, name: string) {
        const types = Helpers.GetTypes(name);
        return `import {\n    ${types.join(',\n    ')}\n} from './${namespace}.${name}';\n\n`;
    }

    public CreateEntity(entityType: OdataEntityType | OdataComplexType) {
        let fileContent = "";
        var imports: string[] = [];
        entityType.navigationproperties.forEach(navProp => {
            const _import = this.CreateImport(navProp.parsedType.namespace, navProp.parsedType.name);
            if (imports.includes(_import)) return;
            fileContent += _import;
            imports.push(_import);
        });
        entityType.properties.forEach(prop => {
            if (prop.parsedType.isComplex) {
                const _import = this.CreateImport(prop.parsedType.namespace, prop.parsedType.name);
                if (imports.includes(_import)) return;
                fileContent += _import;
                imports.push(_import);
            }

        });
        fileContent += this.CreateModel(entityType);
        return fileContent;
    }
}

